package org.example.sensor.concurrent;

/*
 * This is free and unencumbered software released into the public domain.
 * Anyone is free to copy, modify, publish, use, compile, sell, or distribute this software, 
 * either in source code form or as a compiled binary, for any purpose, commercial or 
 * non-commercial, and by any means.
 * 
 * In jurisdictions that recognize copyright laws, the author or authors of this 
 * software dedicate any and all copyright interest in the software to the public domain. 
 * We make this dedication for the benefit of the public at large and to the detriment of 
 * our heirs and successors. We intend this dedication to be an overt act of relinquishment in 
 * perpetuity of all present and future rights to this software under copyright law.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES 
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * For more information, please refer to: https://unlicense.org/
 */

import java.time.Instant;
import java.util.Random;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.commons.lang3.RandomStringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.example.sensor.Sensor;
import org.example.sensor.SensorReporter;
import org.example.sensor.SensorStatus;
import org.example.sensor.data.SensorData;
import org.example.sensor.data.SensorData.SensorType;
import org.example.sensor.simulator.SensorSimulator;

/**
 * This is the <strong>concurrent</strong> implementation of the Sensor interface.
 * <br>
 * <br>
 * This will simulate a IOT Sensor, by:
 * <ul>
 *   <li>Generating a SensorData record every 3 seconds while the Sensor is RUNNING</li>
 *   <li>Pushing the SensorData will be sent at 3 seconds intervals to the SensorRecorder</li>
 * </ul>
 * <br>
 * <br>
 * The initial status of a newly constructed SensorSimulator is INITIALIZED.
 * 
 * @author Jonathan Earl
 * @since 1.0
 * @version 1.0
 * 
 */
public final class SensorImpl 
	implements Sensor
{
	private static final Logger LOG = LogManager.getLogger();

	private static final int INTERVAL = 3000; 

	private SensorStatus status;
	private SensorData sensorData;
	private SensorSimulator simulator;
	private SensorReporter recorder;
	
	private Worker worker;
	private Thread runner;

	private double sensorValue;
	private Lock statusLock;

	/**
	 * The default constructor for the <strong>sequential</strong> SensorImpl.
	 * <br>
	 * <br>
	 * The initial values are:
	 * <ul>
	 *   <li>SensorStatus: <strong>INITIALIZED</strong></li>
	 *   <li>sensorData: 
	 *     <ul>
	 *       <li>SensorType: <strong>a random SensorType</strong></li>
	 *       <li>Value: <strong>a random double</strong></li>
	 *       <li>TimeStamp: <strong>Current Instant</strong></li>
	 *       <li>SensorId: <strong>SensorType concatenated w/ 3 random uppercase letters</strong></li>
	 *     </ul>
	 *   </li>
	 * </ul>
	 * 
	 * @param simulator the SensorSimulator to use as a data generator
	 * @throws InterruptedException 
	 * @throws IllegalArgumentException is the simulator is null
	 */
	public SensorImpl(final SensorSimulator simulator) throws InterruptedException
	{
		LOG.debug("constructing sequential SensorImpl");
		if (simulator == null)
		{
			LOG.error("the SensorSimulator is null");
			throw new IllegalArgumentException("SensorSimulator cannot be null");
		}
		worker = new Worker();
		this.simulator = simulator;
		this.statusLock = new ReentrantLock();
		initializeSensorData();
		setStatus(SensorStatus.INITIALIZED);
	}

	private void initializeSensorData()
	{
		LOG.debug("initializing SersorData");
		final int max = 3;
		Random rand = new Random();
		int selection = rand.nextInt(max);
		SensorType sensorType = SensorType.DEFAULT;
		switch (selection)
		{
		case 0:
			sensorType = SensorType.TEMPERATURE;
			break;
		case 1:
			sensorType = SensorType.VOLUME;
			break;
		case 2:
			sensorType = SensorType.PRESSURE;
			break;
		}
		Instant sensorTimestamp = Instant.now();
		sensorValue = simulator.initialValue(sensorType);
		String sensorId = sensorType.toString() 
				+ "-" + RandomStringUtils.randomAlphabetic(3).toUpperCase();
		LOG.info("SensorId: " + sensorId);
		sensorData = new SensorData(sensorType, sensorTimestamp, sensorValue, sensorId);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setReporter(final SensorReporter recorder) throws InterruptedException 
	{
		LOG.debug("setting the SensorReporter: " + recorder);
		if (recorder == null)
		{
			LOG.error("the SensorReporter is null");
			throw new IllegalArgumentException("SensorReporter cannot be null");
		}
		if (getStatus() != SensorStatus.INITIALIZED)
		{
			LOG.error("Sensor is not INITIALIZED");
			throw new IllegalStateException("Sensor is not INITIALIZED");
		}
		this.recorder = recorder;
	}	

	/**
	 * {@inheritDoc}
	 */
	@Override
	public SensorStatus getStatus() throws InterruptedException
	{
		LOG.debug("returning SersorStatus: " + status);
		statusLock.lock();
		SensorStatus temp = status;
		statusLock.unlock();
		return temp;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setStatus(final SensorStatus status) throws InterruptedException
	{
		LOG.debug("setting SersorStatus: " + status);
		statusLock.lock();
		this.status = status;
		statusLock.unlock();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void startReporting() throws InterruptedException
	{
		LOG.debug("starting to report sensor data");
		if (recorder == null)
		{
			LOG.error("the SensorReporter is null");
			throw new IllegalArgumentException("SensorReporter is null");
		}
		SensorStatus temp = getStatus();
		if (temp == SensorStatus.INITIALIZED || temp == SensorStatus.STOPPED)
		{
			setStatus(SensorStatus.RUNNING);
			
			runner = new Thread(worker);
			runner.setName("Sensor Worker");
			runner.start();
		}
		else
		{
			LOG.error("Sensor is not INITIALIZED or STOPPED");
			throw new IllegalStateException("Sensor is not INITIALIZED or STOPPED");
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void stopReporting() throws InterruptedException
	{
		LOG.debug("stopping the reporting sensor data");
		if (getStatus() != SensorStatus.RUNNING)
		{
			LOG.error("Sensor is not RUNNING");
			throw new IllegalStateException("Sensor is not RUNNING");
		}
		setStatus(SensorStatus.STOPPED);
		runner.interrupt();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void shutdown() throws InterruptedException
	{
		LOG.debug("Shutting down the sensor");
		if (getStatus() == SensorStatus.SHUTDOWN)
		{
			LOG.error("Sensor is already SHUTDOWN");
			throw new IllegalStateException("Sensor is already SHUTDOWN");
		}
		if (getStatus() == SensorStatus.STOPPED || getStatus() == SensorStatus.INITIALIZED)
		{
			setStatus(SensorStatus.SHUTDOWN);
		}
		else
		{
			LOG.error("Sensor is not INITIALIZED or STOPPED");
			throw new IllegalStateException("Sensor is not INITIALIZED or STOPPED");
		}		
	}

	private class Worker
		implements Runnable
	{
		@Override
		public void run()
		{
			LOG.debug("running the sensor");
			while (true)
			{
				try
				{
					if (getStatus() != SensorStatus.RUNNING)
					{
						LOG.trace("Ending sensor run");
						return;
					}
				} 
				catch (InterruptedException e1)
				{
					LOG.warn("Thread was interrupted when checking lock");
					throw new IllegalStateException("Thread was interrupted when checking lock");
				}
				try
				{
					Thread.sleep(INTERVAL);
				} 
				catch (InterruptedException e)
				{
					LOG.warn("Sleep was interrupted");
				}

				SensorType sensorType = sensorData.getType();
				String sensorId = sensorData.getId();
				double currentValue = sensorData.getValue();

				sensorValue = simulator.getNewValue(sensorType, currentValue);
				Instant sensorTimestamp = Instant.now();
				sensorData = new SensorData(sensorType, sensorTimestamp, sensorValue, sensorId);

				recorder.pushSensorData(sensorData);
			}
		}

	}
}
